/*------------------------------------------------------------------------------
* convkml.c : convert solutions to google earth kml file
*
*          Copyright (C) 2007-2009 by T.TAKASU, All rights reserved.
*
* version : $Revision: 1.1 $ $Date: 2008/07/17 21:48:06 $
* history : 2007/01/20  1.0 new
*           2007/03/15  1.1 modify color sequence
*           2007/04/03  1.2 add geodetic height option
*                           support input of NMEA GGA sentence
*                           delete altitude info for track
*                           add time stamp option
*                           separate readsol.c file
*           2009/01/19  1.3 fix bug on display mark with by-q-flag option
*-----------------------------------------------------------------------------*/
#include "rtklib.h"

static const char rcsid[]="$Id: convkml.c,v 1.1 2008/07/17 21:48:06 ttaka Exp $";

/* constants -----------------------------------------------------------------*/

#define SIZP     0.2            /* mark size of rover positions */
#define SIZR     0.3            /* mark size of reference position */
#define TINT     30.0           /* time label interval (sec) */
#define HEADKML1 "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
#define HEADKML2 "<kml xmlns=\"http://earth.google.com/kml/2.1\">"
#define MARKICON "http://maps.google.com/mapfiles/kml/pal2/icon18.png"

/* output track --------------------------------------------------------------*/
static void outtrack(FILE *f, const solp_t *sol, const double *mpos,
                     const char *color, int outalt, int outtime)
{
    int i;
    double alt=0.0,p[2];
    
    fprintf(f,"<Placemark>\n");
    fprintf(f,"<name>Rover Track</name>\n");
    fprintf(f,"<Style>\n");
    fprintf(f,"<LineStyle>\n");
    fprintf(f,"<color>%s</color>\n",color);
    fprintf(f,"</LineStyle>\n");
    fprintf(f,"</Style>\n");
    fprintf(f,"<LineString>\n");
    if (outalt) fprintf(f,"<altitudeMode>absolute</altitudeMode>\n");
    fprintf(f,"<coordinates>\n");
    for (i=0;i<sol->n;i++) {
        if (outalt) {
            p[0]=sol->data[i].pos[0]*D2R; p[1]=sol->data[i].pos[1]*D2R;
            alt=sol->data[i].pos[2]-(outalt==2?geoidh(p):0.0);
        }
        fprintf(f,"%13.9f,%12.9f,%5.3f\n",
                sol->data[i].pos[1],sol->data[i].pos[0],alt);
    }
    fprintf(f,"</coordinates>\n");
    fprintf(f,"</LineString>\n");
    fprintf(f,"</Placemark>\n");
}
/* output point --------------------------------------------------------------*/
static void outpoint(FILE *fp, gtime_t time, const double *pos,
                     const char *label, int style, int outalt, int outtime)
{
    char str[256]="";
    double alt=0.0,p[2],ep[6];
    
    fprintf(fp,"<Placemark>\n");
    if (*label) fprintf(fp,"<name>%s</name>\n",label);
    fprintf(fp,"<styleUrl>#P%d</styleUrl>\n",style);
    if (outtime) {
        if      (outtime==2) time=gpst2utc(time);
        else if (outtime==3) time=timeadd(gpst2utc(time),9*3600.0);
        time2epoch(time,ep);
        if (!*label&&fmod(ep[5]+0.005,TINT)<0.01) {
            sprintf(str,"%02.0f:%02.0f:%02.0f",ep[3],ep[4],ep[5]);
            fprintf(fp,"<name>%s</name>\n",str);
        }
        sprintf(str,"%04.0f-%02.0f-%02.0fT%02.0f:%02.0f:%05.2fZ",
                ep[0],ep[1],ep[2],ep[3],ep[4],ep[5]);
        fprintf(fp,"<TimeStamp><when>%s</when></TimeStamp>\n",str);
    }
    fprintf(fp,"<Point>\n");
    if (outalt) {
        fprintf(fp,"<extrude>1</extrude>\n");
        fprintf(fp,"<altitudeMode>absolute</altitudeMode>\n");
        p[0]=pos[0]*D2R; p[1]=pos[1]*D2R;
        alt=pos[2]-(outalt==2?geoidh(p):0.0);
    }
    fprintf(fp,"<coordinates>%13.9f,%12.9f,%5.3f</coordinates>\n",pos[1],
            pos[0],alt);
    fprintf(fp,"</Point>\n");
    fprintf(fp,"</Placemark>\n");
}
/* save kml file -------------------------------------------------------------*/
static int savekml(const char *file, const solp_t *sol, double rpos[][3],
                   int nrpos, const double *mpos, int tcolor, int pcolor,
                   int outalt, int outtime)
{
    FILE *fp;
    char *color[]={"ffffffff","ff008800","ff00aaff","ff0000ff","ff00ffff"};
    int i,qcolor[]={0,1,2,0,4,3,0};
    
    if (!(fp=fopen(file,"w"))) {
        fprintf(stderr,"file open error : %s\n",file);
        return 0;
    }
    fprintf(fp,"%s\n%s\n",HEADKML1,HEADKML2);
    fprintf(fp,"<Document>\n");
    for (i=0;i<5;i++) {
        fprintf(fp,"<Style id=\"P%d\">\n",i);
        fprintf(fp,"  <IconStyle>\n");
        fprintf(fp,"    <color>%s</color>\n",color[i]);
        fprintf(fp,"    <scale>%.1f</scale>\n",i==0?SIZR:SIZP);
        fprintf(fp,"    <Icon><href>%s</href></Icon>\n",MARKICON);
        fprintf(fp,"  </IconStyle>\n");
        fprintf(fp,"</Style>\n");
    }
    if (tcolor>0) {
        outtrack(fp,sol,mpos,color[tcolor-1],outalt,outtime);
    }
    if (pcolor>0) {
        fprintf(fp,"<Folder>\n");
        fprintf(fp,"  <name>Rover Position</name>\n");
        for (i=0;i<sol->n;i++) {
            outpoint(fp,sol->data[i].time,sol->data[i].pos,"",
                     pcolor==5?qcolor[sol->data[i].posq]:pcolor-1,outalt,outtime);
        }
        fprintf(fp,"</Folder>\n");
    }
    for (i=0;i<nrpos;i++) {
        if (rpos[i][0]==0.0||rpos[i][2]<-100.0) continue;
        outpoint(fp,sol->data[0].time,rpos[i],"Reference Position",0,outalt,0);
    }
    fprintf(fp,"</Document>\n");
    fprintf(fp,"</kml>\n");
    fclose(fp);
    return 1;
}
/* convert to google earth kml file --------------------------------------------
* convert position file to google earth kml file
* post-processing positioning
* args   : char   *infile   I   input solution file
*          char   *outfile  I   output solution file ("":<infile>.kml)
*          gtime_t ts,te    I   start/end time
*          int    tint      I   time interval (s) (0.0:all)
*          int    qflg      I   quality flag (0:all)
*          double *offset   I   add offset {east,north,up} (m)
*          int    tcolor    I   track color
*                               (0:none,1:white,2:green,3:orange,4:red,5:yellow)
*          int    pcolor    I   point color
*                               (0:none,1:white,2:green,3:orange,4:red,5:by qflag)
*          int    outalt    I   output altitude (0:off,1:elipsoidal,2:geodetic)
*          int    outtime   I   output time (0:off,1:gpst,2:utc,3:jst)
* return : status (0:ok,-1:file read,-2:file format,-3:no data,-4:file write)
*-----------------------------------------------------------------------------*/
extern int convkml(const char *infile, const char *outfile, gtime_t ts,
                   gtime_t te, double tint, int qflg, double *offset,
                   int tcolor, int pcolor, int outalt, int outtime)
{
    solp_t sol={0};
    int i,j,nrpos,posf;
    char *p,file[1024];
    double rpos[16][3]={{0}},mpos[3]={0.0},off[3];
    
    if (!*outfile) {
        if ((p=strrchr(infile,'.'))) {
            strncpy(file,infile,p-infile);
            strcpy(file+(p-infile),".kml");
        }
        else sprintf(file,"%s.kml",infile);
    }
    else strcpy(file,outfile);
    
    /* read solution file */
    posf=readsolt((char **)&infile,1,ts,te,tint,qflg,&sol,rpos,&nrpos,NULL,NULL);
    if (posf<0) return -1; if (posf>=2) return -2; if (sol.n<=0) return -3;
    
    /* mean position */
    for (i=0;i<3;i++) {
        for (j=0;j<sol.n;j++) mpos[i]+=sol.data[j].pos[i];
        mpos[i]/=sol.n;
    }
    /* add offset */
    off[0]=offset[1]/RE*R2D;
    off[1]=offset[0]/RE*R2D*cos(mpos[0]*D2R);
    off[2]=offset[2];
    for (i=0;i<sol.n;i++) {
        for (j=0;j<3;j++) sol.data[i].pos[j]+=off[j];
    }
    if (rpos[0][0]!=0.0) {
        for (j=0;j<3;j++) rpos[0][j]+=off[j];
    }
    /* save kml file */
    return savekml(file,&sol,rpos,nrpos,mpos,tcolor,pcolor,outalt,outtime)?0:-4;
}
